"""Utilities for working with SLURM."""

import os

from pandas import DataFrame


def get_subject(bids_table: DataFrame) -> str:
    """Get subject from BIDSTable and slurm array task id."""
    array_id = os.getenv("SLURM_ARRAY_TASK_ID")
    subjects = bids_table.flat["sub"].sort_values().unique().tolist()

    if array_id:
        return subjects[int(array_id)]
    elif len(subjects) > 0:
        print("No array id provided - using first subject found")
        return subjects[0]
    else:
        raise ValueError("No array id found")



"""Utility functions for working with BIDS-associated objects."""

from typing import Literal, overload

import bids2table as b2t
import pyarrow as pa
import pyarrow.parquet as pq
from bids2table._pathlib import PathT, as_path
from pandas import DataFrame

from niwrap_helper.types import StrPath


@overload
def get_bids_table(
    dataset_dir: StrPath, index: StrPath, return_type: Literal["pandas"]
) -> DataFrame: ...


@overload
def get_bids_table(
    dataset_dir: StrPath, index: StrPath, return_type: Literal["pyarrow"]
) -> pa.Table: ...


def get_bids_table(
    dataset_dir: StrPath,
    index: StrPath | None = ".index.b2t",
    return_type: Literal["pandas"] | Literal["pyarrow"] = "pandas",
) -> pa.Table | DataFrame:
    """Get and return BIDSTable for a given dataset."""
    dataset_dir = as_path(dataset_dir)

    # Get table
    if index is not None and (index_fp := (dataset_dir / index)).exists():
        table = pq.read_table(index_fp)
    else:
        tables = b2t.batch_index_dataset(b2t.find_bids_datasets(dataset_dir))  # type: ignore
        table = pa.concat_tables(tables)

    # Set each extra entity as separate column
    extra_entities = table.column("extra_entities").to_pylist()
    extra_entities_dicts = [
        dict(pairs) if isinstance(pairs, list) else {} for pairs in extra_entities
    ]
    all_keys = set().union(*(d.keys() for d in extra_entities_dicts if d))

    if all_keys:
        extra_entities_dicts = [
            {k: d.get(k) for k in all_keys} for d in extra_entities_dicts
        ]
        extra_entities_table = pa.Table.from_pylist(extra_entities_dicts)
        table = pa.concat_tables(
            [table, extra_entities_table], promote_options="default"
        )
    table = table.drop(["extra_entities"])

    dispatch = {"pandas": table.to_pandas, "pyarrow": lambda: table}
    try:
        return dispatch[return_type]()
    except KeyError:
        raise ValueError(f"Unsupported return_type: {return_type}")


@overload
def bids_path(
    directory: Literal[False], return_path: Literal[False], **entities
) -> str: ...


@overload
def bids_path(
    directory: Literal[True], return_path: Literal[False], **entities
) -> PathT: ...


@overload
def bids_path(
    directory: Literal[False], return_path: Literal[True], **entities
) -> PathT: ...


def bids_path(
    directory: bool = False, return_path: bool = False, **entities
) -> StrPath:
    """Generate BIDS name / path."""
    if directory and return_path:
        raise ValueError("Only one of 'directory' or 'return_path' can be True")
    name = b2t.format_bids_path(entities)
    return name.parent if directory else name if return_path else name.name



"""Styx-related functions."""

import logging
import os
import shutil
from pathlib import Path
from typing import overload

import yaml
from styxdefs import LocalRunner, OutputPathType, set_global_runner
from styxdocker import DockerRunner
from styxsingularity import SingularityRunner

from niwrap_helper.types import (
    DockerType,
    LocalType,
    SingularityType,
    StrPath,
    StyxRunner,
)


@overload
def setup_styx(
    image_map: StrPath | None,
    runner: DockerType,
    tmp_env: str = "LOCAL",
    tmp_dir: str = "styx_tmp",
) -> tuple[logging.Logger, DockerRunner]: ...


@overload
def setup_styx(
    image_map: StrPath | None,
    runner: SingularityType,
    tmp_env: str = "LOCAL",
    tmp_dir: str = "styx_tmp",
) -> tuple[logging.Logger, SingularityRunner]: ...


@overload
def setup_styx(
    image_map: StrPath | None,
    runner: LocalType,
    tmp_env: str = "LOCAL",
    tmp_dir: str = "styx_tmp",
) -> tuple[logging.Logger, LocalRunner]: ...


def setup_styx(
    image_map: StrPath | None = None,
    runner: str = "local",
    tmp_env: str = "LOCAL",
    tmp_dir: str = "styx_tmp",
) -> tuple[logging.Logger, StyxRunner]:
    """Setup Styx runner."""
    runner = runner.lower()
    if runner == "docker":
        styx_runner = DockerRunner()
    elif runner == "singularity" or runner == "apptainer":
        if image_map is None:
            raise ValueError("No container mapping provided")
        styx_runner = SingularityRunner(
            images=yaml.safe_load(Path(image_map).read_text())
        )
    else:
        styx_runner = LocalRunner()
    styx_runner.data_dir = Path(os.getenv(tmp_env, "/tmp")) / tmp_dir

    set_global_runner(styx_runner)

    return logging.getLogger(styx_runner.logger_name), styx_runner


def gen_hash(runner: StyxRunner) -> str:
    """Generate hash for styx runner."""
    runner.execution_counter += 1
    return f"{runner.uid}_{runner.execution_counter - 1}"


def cleanup(runner: StyxRunner) -> None:
    """Clean up after completing run."""
    shutil.rmtree(runner.data_dir)
    runner.execution_counter = 0


def save(
    files: OutputPathType | list[OutputPathType],
    out_dir: Path,
) -> None:
    """Save file(s) to specified output directory, preserving directory structure."""

    def _save_file(fpath: Path) -> None:
        """Save individual file, preserving directory structure."""
        for part in fpath.parts:
            if part.startswith("sub-"):
                out_fpath = out_dir.joinpath(*fpath.parts[fpath.parts.index(part) :])
                out_fpath.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy2(fpath, out_fpath)
                return
        raise ValueError(f"Unable to find relevant file path components for {fpath}")

    # Ensure `files` is iterable and process each one
    for file in [files] if isinstance(files, (str, Path)) else files:
        _save_file(Path(file))



"""Custom types."""

from typing import Literal, TypeAlias

from bids2table._pathlib import PathT
from styxdefs import LocalRunner
from styxdocker import DockerRunner
from styxsingularity import SingularityRunner

StrPath = str | PathT
StyxRunner = LocalRunner | DockerRunner | SingularityRunner


DockerType: TypeAlias = Literal["docker", "Docker", "DOCKER"]
SingularityType: TypeAlias = Literal[
    "singularity", "Singularity", "SINGULARITY", "apptainer", "Apptainer", "APPTAINER"
]
LocalType: TypeAlias = Literal["local", "Local", "LOCAL"]
